/*
Extracts the "ip" field from Project Sonar data files.
https://scans.io/study/sonar.http
https://scans.io/study/sonar.https
Usage: ./sonar-ips 20160830-http.gz
*/

package main

import (
	"bufio"
	"encoding/json"
	"flag"
	"fmt"
	"io"
	"os"
	"os/exec"
	"runtime"
	"strings"
)

var lineChan chan []byte

func openGzip(filename string) (io.ReadCloser, error) {
	cmd := exec.Command("gzip", "-d", "-c", filename)
	stdout, err := cmd.StdoutPipe()
	if err != nil {
		return nil, err
	}
	err = cmd.Start()
	return stdout, err
}

type Entry struct {
	IP   string `json:"ip"`
}

func processLine(line []byte) error {
	var entry Entry
	err := json.Unmarshal(line, &entry)
	if err != nil {
		return err
	}
	fmt.Println(entry.IP)
	return nil
}

// processFromChan reads from lineChan and feeds them to processLine.
// panics on any error.
func processFromChan() {
	for line := range lineChan {
		err := processLine(line)
		if err != nil {
			fmt.Fprintln(os.Stderr, err)
			os.Exit(1)
		}
	}
}

// processInput reads lines and passes them into lineChan.
func processInput(r io.Reader) error {
	s := bufio.NewScanner(r)
	s.Buffer(nil, 1024*1024)

	for s.Scan() {
		line := s.Bytes()
		// Make a copy because the memory used by s.Bytes may be
		// overwritten.
		lineCopy := make([]byte, len(line))
		copy(lineCopy, line)
		lineChan <- lineCopy
	}

	return s.Err()
}

func processInputFile(filename string) error {
	var f io.ReadCloser
	var err error

	if strings.HasSuffix(filename, ".gz") {
		f, err = openGzip(filename)
	} else {
		f, err = os.Open(filename)
	}
	if err != nil {
		return err
	}
	defer f.Close()

	return processInput(f)
}

func main() {
	flag.Parse()

	lineChan = make(chan []byte, 100)
	defer func() { close(lineChan) }()

	for i := 0; i < runtime.GOMAXPROCS(0); i++ {
		go processFromChan()
	}

	for _, filename := range flag.Args() {
		err := processInputFile(filename)
		if err != nil {
			fmt.Fprintln(os.Stderr, err)
			os.Exit(1)
		}
	}
}
